@@ -158,7 +158,7 @@ sub build {
 }
 
 # authenticator required only for password attribute
-# av:  {Name,Value} or {Id,Type,VendorId,Value}
+# av:  {Name,Value,[Tag]} or {Id,Type,Value,[VendorId],[Tag]}
 sub pack_attribute {
     my ($self, $av, $authenticator) = @_;
 
@@ -168,13 +168,20 @@ sub pack_attribute {
     my $attr;
     my $vendor_id;
 
-    # attribute not present in dictionary must be passed as {Id, Type, VendorId, Value}, where VendorId is optional
+    # attribute not present in dictionary must be passed as {Id, Type, Value, VendorId, Tag },
+    # where VendorId and Tag are optional
     if ($av->{Id}) {
         if (! $av->{Type}) {
             warn "No attribute type for " . $av->{Id};
             return undef;
         }
-        $attr = { id => $av->{Id}, name => $av->{Id}, type => $av->{Type}, vendor => $av->{VendorId} };
+        $attr = {
+            id => $av->{Id},
+            name => $av->{Id},
+            type => $av->{Type},
+            vendor => $av->{VendorId},
+            has_tag => defined $av->{Tag},
+        };
         $vendor_id = $av->{VendorId};
     }
     else {
@@ -185,16 +192,39 @@ sub pack_attribute {
             return undef;
         }
 
+        # tagged attribute
+        if ($av->{Name} =~ /^([\w-]+):(\d+)$/) {
+            ($av->{Name}, $av->{Tag}) = ($1, $2);
+        }
+
         $attr = $dict->attribute($av->{Name});
         if (! $attr) {
             warn "Unknown attribute ".$av->{Name};
             return undef;
         }
 
+        if (defined $av->{Tag} && !$attr->{has_tag}) {
+            warn "Tag not required for attribute ".$av->{Name};
+            return undef;
+        }
+
+        if ($attr->{has_tag} && ! defined $av->{Tag}) {
+            warn "No Tag provided for attribute ".$av->{Name};
+            return undef;
+        }
+
         # TODO store vendor_id in dictionary parser
         $vendor_id = $dict->vendor_id($attr->{vendor});
     }
 
+    if (defined $av->{Tag}) {
+        my $tag = $av->{Tag} // 0;
+        if ($tag < 1 || $tag > 31) {
+            warn "Tag value is out of range [1..31] for ".($av->{Name} // $av->{Id});
+            return undef;
+        }
+    }
+
     my $value = $av->{Value};
     if (! defined $value) {
         warn "Undefined value for " . $attr->{name};
@@ -222,6 +252,10 @@ sub pack_attribute {
     my $len_encoded = length($encoded);
 
     if (! $vendor_id) {
+        if ($attr->{has_tag}) {
+            return pack('C C C', $attr->{id}, $len_encoded + 3, $av->{Tag}) . $encoded;
+        }
+
         return pack('C C', $attr->{id}, $len_encoded + 2) . $encoded;
     }
 
@@ -232,7 +266,12 @@ sub pack_attribute {
         $vsa_header = pack('N C C C', $vendor_id, $attr->{id}, $len_encoded + 3, 0);
     }
     else {
-        $vsa_header = pack('N C C', $vendor_id, $attr->{id}, $len_encoded + 2);
+        if ($attr->{has_tag}) {
+            $vsa_header = pack('N C C C', $vendor_id, $attr->{id}, $len_encoded + 3, $av->{Tag});
+        }
+        else {
+            $vsa_header = pack('N C C', $vendor_id, $attr->{id}, $len_encoded + 2);
+        }
     }
 
     return pack('C C', ATTR_VENDOR, length($vsa_header) + $len_encoded + 2) . $vsa_header . $encoded;
@@ -285,7 +324,7 @@ sub parse {
     my $len = length($attributes);
 
     while ($pos < $len) {
-        my ($attr_val, $vendor_id, $vendor, $vsa_len) = ();
+        my ($attr_val, $vendor_id, $vendor, $vsa_len, $attr, $tag) = ();
         # FIXME not supported
         my $wimax_cont;
 
@@ -303,11 +342,31 @@ sub parse {
                 printf "WIMAX cont: %d\n", $wimax_cont;
             }
 
-            $attr_val = substr($attributes, $pos + 2 + $vsa_header_len, $attr_len - 2 - $vsa_header_len);
-            $vendor = $self->dict()->vendor_name($vendor_id) // $vendor_id;
+            if ($dict) {
+                $vendor = $dict->vendor_name($vendor_id) // $vendor_id;
+                $attr = $dict->attribute_name($vendor, $attr_id);
+            }
+
+            if ($attr && $attr->{has_tag}) {
+                $tag = unpack('C', substr($attributes, $pos + 2 + $vsa_header_len, 1));
+                $attr_val = substr($attributes, $pos + 3 + $vsa_header_len, $attr_len - 3 - $vsa_header_len);
+            }
+            else {
+                $attr_val = substr($attributes, $pos + 2 + $vsa_header_len, $attr_len - 2 - $vsa_header_len);
+            }
         }
         else {
-            $attr_val = substr($attributes, $pos + 2, $attr_len - 2);
+            if ($dict) {
+                $attr = $dict->attribute_name(undef, $attr_id);
+            }
+
+            if ($attr && $attr->{has_tag}) {
+                $tag = unpack('C', substr($attributes, $pos + 2, 1));
+                $attr_val = substr($attributes, $pos + 3, $attr_len - 3);
+            }
+            else {
+                $attr_val = substr($attributes, $pos + 2, $attr_len - 2);
+            }
         }
 
         if ($attr_id == ATTR_MSG_AUTH && ! $vendor) {
@@ -318,8 +377,6 @@ sub parse {
             substr($attributes, $pos + 2, $attr_len - 2, $attr_val);
         }
 
-        my $attr = $dict ? $dict->attribute_name($vendor, $attr_id) : undef;
-
         $pos += $attr_len;
 
         if (! $attr) {
@@ -329,6 +386,7 @@ sub parse {
                 Value => $attr_val,
                 Type => undef,
                 Vendor => $vendor,
+                Tag => undef,
             };
             next;
         }
@@ -349,6 +407,7 @@ sub parse {
             Value => $decoded,
             Type => $attr->{type},
             Vendor => $vendor,
+            Tag => $tag,
         };
     }
 
@@ -443,7 +502,7 @@ Secret is global secret string, can be overrided when building a new packet.
 
 =over
 
-=item build ( type => CODE, av_list => AVLIST, [ authenticator => AUTH ], 
+=item build ( type => CODE, av_list => AVLIST, [ authenticator => AUTH ],
             [ dict => DICTIONARY ], [ secret => SECRET ],
             [ with_msg_auth => BOOL ], [ request_id => BYTE ])
 
@@ -451,7 +510,8 @@ Build a binary-encoded RADIUS packet.
 
 C<type> identify type of RADIUS request. They are defined in Data::Radius::Constants.
 
-C<av_list> is ARRAY-REF of attributes, each defined as HASH-REF with keys {Name, Value} or {Id, [VendorId], Value}
+C<av_list> is ARRAY-REF of attributes, each defined as HASH-REF with keys {Name, Value, [Tag]} or {Id, [VendorId], Value}
+Tagged attributes can be also specified using 'Name:Tag' format.
 
 C<authenticator> is optional for request (by default the random value will be used), but required for replies.
 
